﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Clients.Hive;
using HeuristicLab.Clients.Hive.Slave.Tests;
using HeuristicLab.Services.Hive.Common.DataTransfer;
using HeuristicLab.Services.Hive.Common.ServiceContracts;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Threading;
using HeuristicLab.Hive;
using HeuristicLab.Services.Hive.Common;

namespace HeuristicLab.Services.Hive.Tests {

  using DT = HeuristicLab.Services.Hive.Common.DataTransfer;
  using System.Diagnostics;

  [TestClass]
  public class ServiceTests {

    [ClassInitialize]
    public static void MyClassInitialize(TestContext testContext) {
      PluginLoader.pluginAssemblies.Any();
      ServiceLocator.Instance = new MockServiceLocator(ServiceLocator.Instance);
    }

    private IHiveService GetLocalService() {
      return new HiveService();
    }

    [TestMethod]
    public void TestJobs() {
      var service = GetLocalService();

      DT.HiveExperiment experiment = new DT.HiveExperiment() {
        Name = "TestExperiment",
        Description = ""
      };

      DT.Job job = new DT.Job() {
        CoresNeeded = 1,
        MemoryNeeded = 0,
        Priority = 0
      };

      DT.JobData jobData = new DT.JobData() {
        Data = PersistenceUtil.Serialize(new MockJob(500, true))
      };

      DT.Plugin plugin1 = new DT.Plugin();
      plugin1.Name = "Tests.MyPlugin";
      plugin1.Version = new Version("1.0.0.0");
      plugin1.UserId = Guid.Empty;
      plugin1.IsLocal = true;
      plugin1.DateCreated = DateTime.Now;

      DT.PluginData pluginData1 = new DT.PluginData();
      pluginData1.PluginId = plugin1.Id;
      pluginData1.FileName = "Tests.MyPlugin-1.0.dll";
      pluginData1.Data = new byte[] { 0, 1, 2, 3, 4, 5 };

      job.PluginsNeededIds.Add(plugin1.Id);

      job.Id = service.AddJob(job, jobData, null);
      experiment.RootJobId = job.Id;

      DT.Job jobLoaded = service.GetJob(job.Id);
      Assert.AreEqual(job.Id, jobLoaded.Id);
      Assert.AreEqual(job.CoresNeeded, jobLoaded.CoresNeeded);
      Assert.AreEqual(job.MemoryNeeded, jobLoaded.MemoryNeeded);
      Assert.AreEqual(job.Priority, jobLoaded.Priority);
      Assert.AreEqual(JobState.Waiting, jobLoaded.JobState);
      Assert.AreEqual(ServiceLocator.Instance.AuthorizationManager.UserId, job.UserId);
      Assert.IsTrue(job.PluginsNeededIds.SequenceEqual(jobLoaded.PluginsNeededIds));

      DT.JobData jobDataLoaded = service.GetJobData(job.Id);
      Assert.AreEqual(job.Id, jobDataLoaded.JobId);
      Assert.IsTrue(jobData.Data.SequenceEqual(jobDataLoaded.Data));

      experiment.Id = service.AddHiveExperiment(experiment);

      DT.HiveExperiment experimentLoaded = service.GetHiveExperiment(experiment.Id);
      Assert.AreEqual(experiment.Id, experimentLoaded.Id);
      Assert.AreEqual(experiment.Name, experimentLoaded.Name);
      Assert.AreEqual(experiment.Description, experimentLoaded.Description);
      Assert.AreEqual(experiment.RootJobId, experimentLoaded.RootJobId);

      service.DeleteHiveExperiment(experiment.Id);
      Assert.AreEqual(null, service.GetHiveExperiment(experiment.Id));
      Assert.AreEqual(null, service.GetJob(job.Id));
      Assert.AreEqual(null, service.GetJobData(job.Id));
    }

    List<DT.Job> jobs = new List<DT.Job>();

    [TestMethod]
    public void TestHeartbeats() {
      var service = GetLocalService();
      // check if group already exists and delete
      var existingSlaveGroup = service.GetSlaveGroups().SingleOrDefault(g => g.Name == "TestGroup");
      if (existingSlaveGroup != null) {
        var slavesToDelete = service.GetSlaves().Where(s => s.ParentResourceId == existingSlaveGroup.Id);
        foreach (var slave in slavesToDelete) service.DeleteSlave(slave.Id);
        service.DeleteSlaveGroup(existingSlaveGroup.Id);
      }

      Guid groupId = service.AddSlaveGroup(new SlaveGroup() { Name = "TestGroup", Description = "Used for unit tests" });

      // create slaves
      var slaves = new List<DT.Slave>();
      for (int i = 0; i < 1; i++) {
        DT.Slave slave = new DT.Slave() {
          Cores = 2,
          Memory = 4096,
          Name = "Slave " + i,
          IsAllowedToCalculate = true,
          SlaveState = SlaveState.Idle,
          CpuSpeed = 2800,
          FreeCores = 2,
          FreeMemory = 3000
        };
        // check if slave with this name already exists and delete
        var existingSlave = service.GetSlaves().Where(s => s.Name == slave.Name).SingleOrDefault();
        if (existingSlave != null) service.DeleteSlave(existingSlave.Id);

        slave.Id = service.AddSlave(slave);
        service.AddResourceToGroup(groupId, slave.Id);
        slaves.Add(slave);
      }

      // create jobs with different group, they should not be assigned
      existingSlaveGroup = service.GetSlaveGroups().SingleOrDefault(g => g.Name == "DummyGroup");
      if (existingSlaveGroup != null) service.DeleteSlaveGroup(existingSlaveGroup.Id);

      Guid dummyGroupId = service.AddSlaveGroup(new SlaveGroup() { Name = "DummyGroup", Description = "Used for unit tests; jobs from this group shall not be calculated" });
      // create dummy jobs
      var dummyJobs = new List<Job>();
      for (int i = 0; i < 2; i++) {
        Job job = new Job() {
          CoresNeeded = 1, MemoryNeeded = 0
        };
        JobData jobData = new JobData() { Data = PersistenceUtil.Serialize(new MockJob(500, false)) };
        job.Id = service.AddJob(job, jobData, new List<Guid> { dummyGroupId });
        dummyJobs.Add(job);
      }

      // create jobs
      for (int i = 0; i < 2; i++) {
        Job job = new Job() {
          CoresNeeded = 1, MemoryNeeded = 0
        };
        JobData jobData = new JobData() { Data = PersistenceUtil.Serialize(new MockJob(500, false)) };
        job.Id = service.AddJob(job, jobData, new List<Guid> { groupId });
        jobs.Add(job);
      }

      // send heartbeats
      foreach (var slave in slaves) {
        new Thread(new ParameterizedThreadStart(RunSlaveThread)).Start(slave);
      }

      IEnumerable<LightweightJob> lightweightJobs;
      do {
        Thread.Sleep(500);
        lightweightJobs = service.GetLightweightJobs(jobs.Select(x => x.Id));
      } while (!lightweightJobs.All(x => x.JobState == JobState.Finished));

      // delete slaves
      foreach (var slave in slaves) {
        service.DeleteSlave(slave.Id);
        Assert.AreEqual(null, service.GetSlave(slave.Id));
      }

      // delete groups
      service.DeleteSlaveGroup(groupId);
      service.DeleteSlaveGroup(dummyGroupId);

      // delete jobs
      foreach (var job in jobs) {
        service.DeleteJob(job.Id);
      }

      // delete dummy jobs
      foreach (var job in dummyJobs) {
        service.DeleteJob(job.Id);
      }
    }

    public void RunSlaveThread(object slaveobj) {
      try {
        var service = GetLocalService();
        Slave slave = (Slave)slaveobj;
        int freeCores = slave.Cores.Value;

        for (int i = 0; i < 10; i++) {

          var messages = service.Heartbeat(new Heartbeat() { SlaveId = slave.Id, FreeMemory = 2423, FreeCores = freeCores, JobProgress = new Dictionary<Guid, TimeSpan>() });
          if (messages.Count == 0) {
            Debug.WriteLine("No job available");
            return; // no more jobs
          }

          Debug.WriteLine("Messages: {0}", string.Join(", ", messages.Select(m => m.Message)));

          Assert.IsTrue(messages.Where(x => x.Message == MessageContainer.MessageType.AbortJob).Count() == 0);
          Assert.IsTrue(messages.Where(x => x.Message == MessageContainer.MessageType.SayHello).Count() == 0);
          Assert.IsTrue(messages.Where(x => x.Message == MessageContainer.MessageType.PauseJob).Count() == 0);

          var calculateJobMessage = messages.Where(x => x.Message == MessageContainer.MessageType.CalculateJob).SingleOrDefault();
          if (calculateJobMessage != null) {
            if (!jobs.Select(j => j.Id).Contains(calculateJobMessage.JobId))
              Assert.Fail("Got job which was not assigned to the slavegroup");

            Debug.WriteLine("Job available, calculating");
            Job job = service.GetJob(calculateJobMessage.JobId);

            JobData jobData = service.GetJobData(job.Id);
            IJob deserializedJob = PersistenceUtil.Deserialize<IJob>(jobData.Data);
            deserializedJob.Start();
            job.JobState = JobState.Finished;
            jobs.Where(x => x.Id == job.Id).Single().JobState = JobState.Finished;
            jobData.Data = PersistenceUtil.Serialize(deserializedJob);
            service.UpdateJob(job, jobData);
            Debug.WriteLine("finished calculating");
          }
        }
      }
      catch (Exception e) {
        Assert.Fail(e.Message, e);
      }
    }
  }
}
